from enum import Enum


class EnvironmentVariableKey(Enum):
    ENVIRONMENT = "ENVIRONMENT"
    APPLICATION_NAME = "APPLICATION_NAME"
    DATABASE_CONNECTION_STRING = "DATABASE_CONNECTION_STRING"
    GITHUB_TOKEN = "GITHUB_TOKEN"
    SUBDOMAIN = "SUBDOMAIN"


class EnvironmentSecretKey:
    pass


class SiriusEnvironmentSecretKey(EnvironmentSecretKey, Enum):
    API_KEY = "API_KEY"
    DISCORD_BOT_TOKEN = "DISCORD_BOT_TOKEN"
    WISE_API_KEY = "WISE_API_KEY"
    IBKR_SERVICE_BASE_URL = "IBKR_SERVICE_BASE_URL"
    ALPHA_VANTAGE_API_KEY = "ALPHA_VANTAGE_API_KEY"
    MASSIVE_API_KEY = "MASSIVE_API_KEY"
    MICROSOFT_CLIENT_ID = "MICROSOFT_CLIENT_ID"
    MICROSOFT_TENANT_ID = "MICROSOFT_TENANT_ID"
    MICROSOFT_CLIENT_SECRET = "MICROSOFT_CLIENT_SECRET"
    MICROSOFT_REDIRECT_URI = "MICROSOFT_REDIRECT_URI"
    LOGIN_REDIRECT_URI = "LOGIN_REDIRECT_URI"
    PRIVATE_KEY = "PRIVATE_KEY"
